from IPython.display import Math, display
from sympy import diff, limit, Mul
import sympy


class Integral(sympy.Integral):

    def by_parts(self, u, vp):
        return integrate_by_parts(self, u, vp)


class PartialDerivative:

    def __init__(self, *wrt):
        self.wrt = wrt

    def __call__(self, f):
        return diff(f, *self.wrt)

    def __str__(self):
        return '\partial_{' + ''.join([str(x) for x in self.wrt]) + '}'

    def __repr__(self):
        return self.__str__()

    def _repr_latex_(self):
        return Math(str(self))

    def _ipython_display_(self, **kwargs):
        display(Math(str(self)))


def integrate_by_parts(integral, u, vp):
    """ Makes SymPy integrate by parts

        integral: A SymPy Integral object
        u: u(x)
        vp: derivative of v(x)
    """
    integrand, (x_, a_, b_) = integral.args

    assert integrand.doit() == (u * vp).doit()
    v = Integral(vp, x_).doit()

    # If the new integrand contains a leading minus sign,
    # put it in front of the integral.
    integrand = diff(u, x_) * v
    sign = -1
    if isinstance(integrand, Mul) and integrand.args[0] == -1:
        integrand *= -1
        sign = 1

    # The limits take into account cases like oo*exp(-oo)
    return (
            limit(u * v, x_, b_) - limit(u * v, x_, a_)
            + sign * Integral(integrand, (x_, a_, b_))
    )


import types
import sys
from .equation import *
from .pick import *
from .calculus import *

from functools import wraps


this_module = sys.modules[__name__]


def _get_imported_names(module):
    names = module.__all__ if hasattr(module, '__all__') else dir(module)
    return [name for name in names if not name.startswith('_')]


def _wrap_function(func):
    @wraps(func)
    def f(*args, **kwargs):
        if isinstance(args[0], Eq):
            if len(args) > 1:
                other_args = tuple(args[1:])
            else:
                other_args = ()
            return args[0].apply('both', func, *other_args, **kwargs)
        else:
            return func(*args, **kwargs)
    return f



_names_from_sympy = _get_imported_names(sympy)


for name in _names_from_sympy:
    obj = getattr(sympy, name)
    if isinstance(obj, types.FunctionType) or isinstance(obj, sympy.FunctionClass):
        setattr(this_module, name, _wrap_function(obj))



from sympy import Wild


def select_by_pattern(expr, pattern, which):
    matches = expr.match(pattern)
    if matches:
        return matches[which]
    return None


def select_atoms(expr, atom_type):
    return list(expr.atoms(atom_type))

a__ = Wild('a')
b__ = Wild('b')
c__ = Wild('c')
d__ = Wild('d')
e__ = Wild('e')
f__ = Wild('f')
g__ = Wild('g')
h__ = Wild('h')
i__ = Wild('i')
j__ = Wild('j')

